// testthread.cpp : main project file.

#include "stdafx.h"
#include <omp.h>
#include <windows.h>


#define C_OUTERLOOP 4
#define C_MIDDLELOOP 20
#define C_INNERLOOP 2000
#define MAXWORKERS 2

struct arguments {
	int i1,i2;
	int innerLoopCount;
};

LARGE_INTEGER GetTickCount2() {
	HANDLE hCurrentProcess = GetCurrentProcess();
	DWORD dwProcessAffinity;
	DWORD dwSystemAffinity;
	LARGE_INTEGER result;

	GetProcessAffinityMask(hCurrentProcess,&dwProcessAffinity, &dwSystemAffinity);
	SetProcessAffinityMask(hCurrentProcess,1);

	QueryPerformanceCounter(&result);

	SetProcessAffinityMask(hCurrentProcess,dwProcessAffinity);

	return result;
}

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------

// hier gibts Arbeit
int work (int loopCount) {
	int result=0;
	for (int i=0; i<loopCount; ++i) 
		result+=i;
	return result;
}

// normale Threads
DWORD WINAPI LoopThreadFunction(LPVOID arguments) {
	struct arguments *args;
	args = (struct arguments *) arguments;
	int innerLoopCount=args->innerLoopCount;
	for (int i=args->i1; i < args->i2; ++i) {
			work(C_INNERLOOP);
	}
	return 0;
}


void testserialCode() {
	for (int i=0; i < C_OUTERLOOP; ++i) {
		for (int j=0; j < C_MIDDLELOOP; ++j) {
			work(C_INNERLOOP);
		}
	}
}

void testSectionCode() {	
	int numThreads = omp_get_num_threads();
	int chunksize=C_MIDDLELOOP/numThreads;
	
	omp_set_num_threads(2);
	for (int i=0; i < C_OUTERLOOP; ++i) {
		#pragma omp parallel sections 
		{
			#pragma omp section
			{
				//int id=omp_get_thread_num();
				//Console::WriteLine(L"Thread id:{0}  i: {1}", id,i);
				for (int j=0; j < chunksize; ++j) {
					work(C_INNERLOOP);
				}
			}
		
			#pragma omp section
			{
				//int id=omp_get_thread_num();
				//Console::WriteLine(L"Thread id:{0}  i: {1}", id,i);
				for (int j=chunksize; j < C_MIDDLELOOP; ++j) {
					work(C_INNERLOOP);
				}
			}
		}
	}

}  // section


void testParallelFor1Code(){
	omp_set_num_threads(1);
	for (int i=0; i < C_OUTERLOOP; ++i) {
		#pragma omp parallel for 
			for (int j=0; j < C_MIDDLELOOP; ++j) {
				//int numThreads = omp_get_num_threads();
				//Console::WriteLine(L"numberThreads: {0} ", numThreads);
				/*if (j==0  || j==11) {
					int id=omp_get_thread_num();
					Console::WriteLine(L"Thread id:{0}  numberThreads: {1} j: {2} ", id,numberThreads, j);
				}
				*/
				work(C_INNERLOOP);
			}

	}
} // testParallelFor1Code

void testParallelForCode(int numberThreads){
	omp_set_num_threads(numberThreads);
	for (int i=0; i < C_OUTERLOOP; ++i) {
		#pragma omp parallel for
			for (int j=0; j < C_MIDDLELOOP; ++j) {
				//int numThreads = omp_get_num_threads();
				//Console::WriteLine(L"numberThreads: {0} ", numThreads);
				/*if (j==0  || j==11) {
					int id=omp_get_thread_num();
					Console::WriteLine(L"Thread id:{0}  j: {1} ", id,j);
				}
				*/				
				work(C_INNERLOOP);
			}

	}
} // testParallelFor1Code


void testParallelFor2Code(){
	omp_set_num_threads(2);
	for (int i=0; i < C_OUTERLOOP; ++i) {
		#pragma omp parallel for
			for (int j=0; j < C_MIDDLELOOP; ++j) {
				//int numThreads = omp_get_num_threads();
				//Console::WriteLine(L"numberThreads: {0} ", numThreads);
				/*if (j==0  || j==11) {
					int id=omp_get_thread_num();
					Console::WriteLine(L"Thread id:{0}  j: {1} ", id,j);
				}
				*/				
				work(C_INNERLOOP);
			}

	}
} // testParallelFor2Code




void testWin32Code(){
	struct arguments args1, args2;
	HANDLE *hThreadVector;
	DWORD ThreadID[MAXWORKERS];
	DWORD CreationFlags =  CREATE_SUSPENDED;

	int stacksize = 10000;
	int numThreads = omp_get_num_threads();
	int chunksize=C_MIDDLELOOP/numThreads;
	args1.i1 = 0;
	args1.i2 = chunksize;
	args1.innerLoopCount = C_INNERLOOP;

	args2.i1 = chunksize;
	args2.i2 = C_MIDDLELOOP;
	args2.innerLoopCount = C_INNERLOOP;

	CreationFlags =  0;
	for (int i=0; i < C_OUTERLOOP; ++i) {
		hThreadVector = new HANDLE[MAXWORKERS];
		hThreadVector[0] = CreateThread (NULL, stacksize, 
			   (LPTHREAD_START_ROUTINE) LoopThreadFunction,
					(LPDWORD ) &args1, CreationFlags, 
					(LPDWORD)&ThreadID[0] );
		hThreadVector[1] = CreateThread (NULL, stacksize, 
			   (LPTHREAD_START_ROUTINE) LoopThreadFunction,
					(LPDWORD ) &args2, CreationFlags, 
					(LPDWORD)&ThreadID[1] );
		//for (int j=0; j<MAXWORKERS; ++j) {
		//	ResumeThread( hThreadVector[j] );
		//} // for
		//free(hThreadVector);
		WaitForMultipleObjects(2,hThreadVector, TRUE, INFINITE);
		CloseHandle(hThreadVector[0]);
		CloseHandle(hThreadVector[1]);
		delete [] hThreadVector;
	}
} // win32 Threads



int _tmain(int argc, _TCHAR* argv[])
{
    LARGE_INTEGER t1,t2;
	long z;

	LARGE_INTEGER Frequency;
	if ( ! QueryPerformanceFrequency( &Frequency ) )
	{
		Frequency.HighPart=0;
		Frequency.LowPart=0;
	}
	printf("Frequency: %u %u\n\n\n ", Frequency.HighPart, Frequency.LowPart );

	// ------------------------------

	t1 = GetTickCount2();
	testserialCode();
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife Seriell time: %u\n ", z );

	t1 = GetTickCount2();
	testSectionCode();	
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife 2+Sectionen  time: %u\n ", z );


	t1 = GetTickCount2();
	testParallelFor1Code();
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife Parallel-For1  time: %u\n ", z );


	t1 = GetTickCount2();
	testParallelFor2Code();
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife Parallel-For2  time: %u\n ", z );


	t1 = GetTickCount2();
	testParallelForCode(1);
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife Parallel-For  time: %u\n ", z );


	t1 = GetTickCount2();
	testParallelForCode(2);
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife Parallel-For  time: %u\n ", z );


	t1 = GetTickCount2();
	testWin32Code();
	t2 = GetTickCount2();
	z=  (t2.HighPart-t1.HighPart)  +  (t2.LowPart-t1.LowPart);
	printf("Schleife Win32Threads  time: %u\n ", z );

	char ch[200];
	gets(ch);

	return 0;
}

